from types import SimpleNamespace
import osqp
import numpy as np
from scipy import sparse

import pytest
import numpy.testing as nptest


@pytest.fixture
def self(algebra, solver_type, atol, rtol, decimal_tol):
    ns = SimpleNamespace()
    ns.P = sparse.triu([[2.0, 5.0], [5.0, 1.0]], format='csc')
    ns.q = np.array([3, 4])
    ns.A = sparse.csc_matrix([[-1.0, 0.0], [0.0, -1.0], [-1.0, 3.0], [2.0, 5.0], [3.0, 4]])
    ns.u = np.array([0.0, 0.0, -15, 100, 80])
    ns.l = -np.inf * np.ones(len(ns.u))
    ns.model = osqp.OSQP(algebra=algebra)
    return ns


def test_non_convex_small_sigma(self, solver_type):
    if solver_type == 'direct':
        with pytest.raises(osqp.OSQPException):
            self.model.setup(
                P=self.P,
                q=self.q,
                A=self.A,
                l=self.l,
                u=self.u,
                solver_type=solver_type,
                sigma=1e-6,
            )
    else:
        self.model.setup(
            P=self.P,
            q=self.q,
            A=self.A,
            l=self.l,
            u=self.u,
            solver_type=solver_type,
            sigma=1e-6,
        )
        res = self.model.solve()

        assert res.info.status_val in (
            self.model.constant('OSQP_MAX_ITER_REACHED'),
            self.model.constant('OSQP_NON_CVX'),
        )


def test_non_convex_big_sigma(self):
    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, sigma=5)
    res = self.model.solve()

    assert res.info.status_val == self.model.constant('OSQP_NON_CVX')


def test_nan(self):
    nptest.assert_approx_equal(self.model.constant('OSQP_NAN'), np.nan)



import os
from types import SimpleNamespace
import osqp
import numpy as np
from scipy import sparse

import pytest


@pytest.fixture
def self(algebra, solver_type, atol, rtol, decimal_tol):
    ns = SimpleNamespace()
    ns.opts = {
        'verbose': False,
        'eps_abs': 1e-05,
        'eps_rel': 1e-05,
        'eps_prim_inf': 1e-15,  # Focus only on dual infeasibility
        'eps_dual_inf': 1e-6,
        'scaling': 3,
        'max_iter': 2500,
        'polishing': False,
        'check_termination': 1,
        'polish_refine_iter': 4,
        'solver_type': solver_type,
    }

    ns.model = osqp.OSQP(algebra=algebra)
    return ns


def test_dual_infeasible_lp(self):
    # Dual infeasible example
    self.P = sparse.csc_matrix((2, 2))
    self.q = np.array([2, -1])
    self.A = sparse.eye(2, format='csc')
    self.l = np.array([0.0, 0.0])
    self.u = np.array([np.inf, np.inf])

    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)

    # Solve problem with OSQP
    res = self.model.solve()

    assert res.info.status_val == self.model.constant('OSQP_DUAL_INFEASIBLE')

    normalized_dual_inf_cert = res.dual_inf_cert / np.linalg.norm(res.dual_inf_cert)
    normalized_dual_inf_cert_correct = np.load(
        os.path.join(os.path.dirname(__file__), 'solutions', 'test_dual_infeasibility.npz')
    )['lp_normalized_dual_inf_cert_correct']
    assert np.allclose(normalized_dual_inf_cert, normalized_dual_inf_cert_correct)


def test_dual_infeasible_qp(self):
    # Dual infeasible example
    self.P = sparse.diags([4.0, 0.0], format='csc')
    self.q = np.array([0, 2])
    self.A = sparse.csc_matrix([[1.0, 1.0], [-1.0, 1.0]])
    self.l = np.array([-np.inf, -np.inf])
    self.u = np.array([2.0, 3.0])

    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)

    # Solve problem with OSQP
    res = self.model.solve()

    assert res.info.status_val == self.model.constant('OSQP_DUAL_INFEASIBLE')

    normalized_dual_inf_cert = res.dual_inf_cert / np.linalg.norm(res.dual_inf_cert)
    normalized_dual_inf_cert_correct = np.load(
        os.path.join(os.path.dirname(__file__), 'solutions', 'test_dual_infeasibility.npz')
    )['qp_normalized_dual_inf_cert_correct']
    assert np.allclose(normalized_dual_inf_cert, normalized_dual_inf_cert_correct)


def test_primal_and_dual_infeasible_problem(self):
    self.n = 2
    self.m = 4
    self.P = sparse.csc_matrix((2, 2))
    self.q = np.array([-1.0, -1.0])
    self.A = sparse.csc_matrix([[1.0, -1.0], [-1.0, 1.0], [1.0, 0.0], [0.0, 1.0]])
    self.l = np.array([1.0, 1.0, 0.0, 0.0])
    self.u = np.inf * np.ones(self.m)

    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)

    # Warm start to avoid infeasibility detection at first step
    x0 = 25.0 * np.ones(self.n)
    y0 = -2.0 * np.ones(self.m)
    self.model.warm_start(x=x0, y=y0)

    # Solve
    res = self.model.solve()

    assert res.info.status_val in (
        self.model.constant('OSQP_PRIMAL_INFEASIBLE'),
        self.model.constant('OSQP_DUAL_INFEASIBLE'),
    )



import os
from osqp import algebra_available


def pytest_generate_tests(metafunc):

    # detect env vars to decide which algebras to include/skip
    algebras_include = os.environ.get('OSQP_TEST_ALGEBRA_INCLUDE', 'builtin mkl-direct mkl-indirect cuda').split()
    algebras_skip = os.environ.get('OSQP_TEST_ALGEBRA_SKIP', '').split()
    algebras = [x for x in algebras_include if x not in algebras_skip]

    parameters = ('algebra', 'solver_type', 'atol', 'rtol', 'decimal_tol')
    values = []
    if algebra_available('builtin') and 'builtin' in algebras:
        values.append(
            ('builtin', 'direct', 1e-3, 1e-4, 4),
        )
    if algebra_available('mkl') and 'mkl-direct' in algebras:
        values.append(
            ('mkl', 'direct', 1e-3, 1e-4, 4),
        )
    if algebra_available('mkl') and 'mkl-indirect' in algebras:
        values.append(
            ('mkl', 'indirect', 1e-3, 1e-4, 3),
        )
    if algebra_available('cuda') and 'cuda' in algebras:
        values.append(
            ('cuda', 'indirect', 1e-2, 1e-3, 2),
        )

    metafunc.parametrize(parameters, values)



import osqp
import numpy as np
from scipy import sparse
import unittest
import pytest
import numpy.testing as nptest
import shutil as sh
import sys


@pytest.mark.skipif(not osqp.algebra_available('builtin'), reason='Builtin Algebra not available')
class codegen_vectors_tests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        P = sparse.diags([11.0, 0.0], format='csc')
        q = np.array([3, 4])
        A = sparse.csc_matrix([[-1, 0], [0, -1], [-1, -3], [2, 5], [3, 4]])
        u = np.array([0, 0, -15, 100, 80])
        l = -np.inf * np.ones(len(u))
        n = P.shape[0]
        m = A.shape[0]
        opts = {
            'verbose': False,
            'eps_abs': 1e-08,
            'eps_rel': 1e-08,
            'rho': 0.01,
            'alpha': 1.6,
            'max_iter': 10000,
            'warm_starting': True,
        }

        model = osqp.OSQP(algebra='builtin')
        if not model.has_capability('OSQP_CAPABILITY_DERIVATIVES'):
            pytest.skip('No derivatives capability')
        model.setup(P=P, q=q, A=A, l=l, u=u, **opts)

        model_dir = model.codegen(
            'codegen_vec_out',
            extension_name='vec_emosqp',
            include_codegen_src=True,
            force_rewrite=True,
            prefix='foo',
            compile=True,
        )
        sys.path.append(model_dir)

        cls.m = m
        cls.n = n
        cls.P = P
        cls.q = q
        cls.A = A
        cls.l = l
        cls.u = u
        cls.opts = opts

    @classmethod
    def tearDownClass(cls):
        sh.rmtree('codegen_vec_out', ignore_errors=True)

    def setUp(self):
        self.model = osqp.OSQP(algebra='builtin')
        self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)

    def test_solve(self):
        import vec_emosqp

        # Solve problem
        x, y, _, _, _ = vec_emosqp.solve()

        # Assert close
        nptest.assert_array_almost_equal(x, np.array([0.0, 5.0]), decimal=5)
        nptest.assert_array_almost_equal(y, np.array([1.66666667, 0.0, 1.33333333, 0.0, 0.0]), decimal=5)

    def test_update_q(self):
        import vec_emosqp

        # Update linear cost and solve the problem
        q_new = np.array([10.0, 20.0])
        vec_emosqp.update_data_vec(q=q_new)
        x, y, _, _, _ = vec_emosqp.solve()

        # Assert close
        nptest.assert_array_almost_equal(x, np.array([0.0, 5.0]), decimal=5)
        nptest.assert_array_almost_equal(y, np.array([3.33333334, 0.0, 6.66666667, 0.0, 0.0]), decimal=5)

        # Update linear cost to the original value
        vec_emosqp.update_data_vec(q=self.q)

    def test_update_l(self):
        import vec_emosqp

        # Update lower bound
        l_new = -100.0 * np.ones(self.m)
        vec_emosqp.update_data_vec(l=l_new)
        x, y, _, _, _ = vec_emosqp.solve()

        # Assert close
        nptest.assert_array_almost_equal(x, np.array([0.0, 5.0]), decimal=5)
        nptest.assert_array_almost_equal(y, np.array([1.66666667, 0.0, 1.33333333, 0.0, 0.0]), decimal=5)

        # Update lower bound to the original value
        vec_emosqp.update_data_vec(l=self.l)

    def test_update_u(self):
        import vec_emosqp

        # Update upper bound
        u_new = 1000.0 * np.ones(self.m)
        vec_emosqp.update_data_vec(u=u_new)
        x, y, _, _, _ = vec_emosqp.solve()

        # Assert close
        nptest.assert_array_almost_equal(x, np.array([-1.51515152e-01, -3.33282828e02]), decimal=4)
        nptest.assert_array_almost_equal(y, np.array([0.0, 0.0, 1.33333333, 0.0, 0.0]), decimal=4)

        # Update upper bound to the original value
        vec_emosqp.update_data_vec(u=self.u)

    def test_update_bounds(self):
        import vec_emosqp

        # Update upper bound
        l_new = -100.0 * np.ones(self.m)
        u_new = 1000.0 * np.ones(self.m)
        vec_emosqp.update_data_vec(l=l_new, u=u_new)
        x, y, _, _, _ = vec_emosqp.solve()

        # Assert close
        nptest.assert_array_almost_equal(x, np.array([-0.12727273, -19.94909091]), decimal=4)
        nptest.assert_array_almost_equal(y, np.array([0.0, 0.0, 0.0, -0.8, 0.0]), decimal=4)

        # Update upper bound to the original value
        vec_emosqp.update_data_vec(l=self.l, u=self.u)



import osqp
from multiprocessing.pool import ThreadPool
import time
import numpy as np
from scipy import sparse
import unittest
import pytest


@pytest.mark.skipif(not osqp.algebra_available('builtin'), reason='Builtin Algebra not available')
class multithread_tests(unittest.TestCase):
    def test_multithread(self):
        data = []

        n_rep = 50

        for i in range(n_rep):
            m = 1000
            n = 500
            Ad = sparse.random(m, n, density=0.3, format='csc')
            b = np.random.randn(m)

            # OSQP data
            P = sparse.block_diag([sparse.csc_matrix((n, n)), sparse.eye(m)], format='csc')
            q = np.zeros(n + m)
            A = sparse.vstack(
                [
                    sparse.hstack([Ad, -sparse.eye(m)]),
                    sparse.hstack([sparse.eye(n), sparse.csc_matrix((n, m))]),
                ],
                format='csc',
            )
            l = np.hstack([b, np.zeros(n)])
            u = np.hstack([b, np.ones(n)])

            data.append((P, q, A, l, u))

        def f(i):
            P, q, A, l, u = data[i]
            m = osqp.OSQP(algebra='builtin')
            m.setup(P, q, A, l, u, verbose=False)
            m.solve()

        pool = ThreadPool(2)

        tic = time.time()
        for i in range(n_rep):
            f(i)
        t_serial = time.time() - tic

        tic = time.time()
        pool.map(f, range(n_rep))
        t_parallel = time.time() - tic

        self.assertLess(t_parallel, t_serial)



from types import SimpleNamespace
import pytest
from scipy import sparse
import numpy as np
import numpy.testing as nptest
from osqp import OSQP
from osqp.tests.utils import load_high_accuracy


@pytest.fixture
def self(algebra, solver_type, atol, rtol, decimal_tol):
    ns = SimpleNamespace()
    ns.P = sparse.diags([11.0, 0.0], format='csc')
    ns.q = np.array([3, 4])
    ns.A = sparse.csc_matrix([[-1, 0], [0, -1], [-1, -3], [2, 5], [3, 4]])
    ns.u = np.array([0.0, 0.0, -15, 100, 80])
    ns.l = -1e06 * np.ones(len(ns.u))
    ns.n = ns.P.shape[0]
    ns.m = ns.A.shape[0]
    ns.opts = {
        'verbose': False,
        'eps_abs': 1e-09,
        'eps_rel': 1e-09,
        'max_iter': 2500,
        'rho': 0.1,
        'adaptive_rho': False,
        'polishing': False,
        'check_termination': 1,
        'warm_starting': True,
        'solver_type': solver_type,
    }
    ns.model = OSQP(algebra=algebra)
    ns.model.setup(P=ns.P, q=ns.q, A=ns.A, l=ns.l, u=ns.u, **ns.opts)
    ns.atol = atol
    ns.rtol = rtol
    ns.decimal_tol = decimal_tol
    return ns


def test_basic_QP(self):
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_basic_QP')
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_q(self):
    # Update linear cost
    q_new = np.array([10, 20])
    self.model.update(q=q_new)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_q')
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_l(self):
    # Update lower bound
    l_new = -50 * np.ones(self.m)
    self.model.update(l=l_new)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_l')
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_u(self):
    # Update lower bound
    u_new = 1000 * np.ones(self.m)
    self.model.update(u=u_new)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_u')
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_bounds(self):
    # Update lower bound
    l_new = -100 * np.ones(self.m)
    # Update lower bound
    u_new = 1000 * np.ones(self.m)
    self.model.update(u=u_new, l=l_new)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_bounds')
    if self.model.algebra != 'cuda':  # pytest-todo
        nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)
    else:
        assert res.info.status_val == self.model.constant('OSQP_PRIMAL_INFEASIBLE')


def test_update_max_iter(self):
    self.model.update_settings(max_iter=80)
    res = self.model.solve()

    assert res.info.status_val == self.model.constant('OSQP_MAX_ITER_REACHED')


def test_update_check_termination(self):
    self.model.update_settings(check_termination=0)
    res = self.model.solve()

    assert res.info.iter == self.opts['max_iter']


def test_update_rho(self):
    res_default = self.model.solve()

    # Setup with different rho and update
    default_opts = self.opts.copy()
    default_opts['rho'] = 0.7
    model = OSQP(algebra=self.model.algebra)
    model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **default_opts)
    model.update_settings(rho=self.opts['rho'])
    res_updated_rho = model.solve()

    # Assert same number of iterations
    assert res_default.info.iter == res_updated_rho.info.iter


def test_upper_triangular_P(self):
    res_default = self.model.solve()

    # Get upper triangular P
    P_triu = sparse.triu(self.P, format='csc')

    # Setup and solve with upper triangular part only
    model = OSQP(algebra=self.model.algebra)
    model.setup(P=P_triu, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)
    res_triu = model.solve()

    nptest.assert_allclose(res_default.x, res_triu.x, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res_default.y, res_triu.y, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(
        res_default.info.obj_val,
        res_triu.info.obj_val,
        decimal=self.decimal_tol,
    )


def test_update_invalid(self):
    # can't update unsupported setting
    with pytest.raises(ValueError):
        self.model.update_settings(foo=42)



import osqp
import numpy as np
import numpy.random as npr
from scipy import sparse
from scipy.optimize import approx_fprime
import numpy.testing as npt
import unittest
import pytest


npr.seed(1)

# Tests settings
grad_precision = 1e-5
rel_tol = 5e-3
abs_tol = 5e-3
rel_tol_relaxed = 1e-2
abs_tol_relaxed = 1e-2

# OSQP settings
eps_abs = 1e-9
eps_rel = 1e-9
max_iter = 500000


@pytest.mark.skipif(not osqp.algebra_available('builtin'), reason='Builtin Algebra not available')
class derivative_tests(unittest.TestCase):
    def setUp(self):
        npr.seed(1)

    def get_prob(self, n=10, m=3, P_scale=1.0, A_scale=1.0):
        L = np.random.randn(n, n - 1)
        # P = sparse.csc_matrix(L.dot(L.T) + 5. * sparse.eye(n))
        P = sparse.csc_matrix(L.dot(L.T) + 0.1 * sparse.eye(n))
        # P = sparse.csc_matrix(L.dot(L.T))
        x_0 = npr.randn(n)
        s_0 = npr.rand(m)
        A = sparse.csc_matrix(npr.randn(m, n))
        u = A.dot(x_0) + s_0
        # l = -10 - 10 * npr.rand(m)
        l = A.dot(x_0) - s_0
        q = npr.randn(n)
        true_x = npr.randn(n)
        true_y = npr.randn(m)

        return [P, q, A, l, u, true_x, true_y]

    def get_grads(self, P, q, A, l, u, true_x, true_y=None):
        # Get gradients by solving with osqp
        m = osqp.OSQP(algebra='builtin')
        m.setup(
            P,
            q,
            A,
            l,
            u,
            eps_abs=eps_abs,
            eps_rel=eps_rel,
            max_iter=max_iter,
            verbose=True,
        )
        results = m.solve()
        if results.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem not solved!')
        x = results.x
        y = results.y
        if true_y is None:
            m.adjoint_derivative_compute(dx=x - true_x)
        else:
            m.adjoint_derivative_compute(dx=x - true_x, dy=y - true_y)

        dP, dA = m.adjoint_derivative_get_mat()
        dq, dl, du = m.adjoint_derivative_get_vec()
        grads = dP, dq, dA, dl, du

        return grads

    def get_forward_grads(self, P, q, A, l, u, dP, dq, dA, dl, du):
        # Get gradients by solving with osqp
        m = osqp.OSQP(algebra='builtin', eps_rel=eps_rel, eps_abs=eps_abs)
        m.setup(
            P,
            q,
            A,
            l,
            u,
            eps_abs=eps_abs,
            eps_rel=eps_rel,
            max_iter=max_iter,
            verbose=False,
        )
        results = m.solve()
        if results.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem not solved!')
        grads = m.forward_derivative(dP=dP, dq=dq, dA=dA, dl=dl, du=du)
        return grads

    @pytest.mark.skip(reason='forward derivatives not implemented yet')
    def test_dsol_dq(self, verbose=False):
        n, m = 5, 5

        prob = self.get_prob(n=n, m=m, P_scale=100.0, A_scale=100.0)
        P, q, A, l, u, true_x, true_y = prob

        def grad(dq):
            [dx, dyl, dyu] = self.get_forward_grads(P, q, A, l, u, None, dq, None, None, None)
            return dx, dyl, dyu

        dq = np.random.normal(size=(n))
        dx_computed, dyl_computed, dyu_computed = grad(dq)

        osqp_solver = osqp.OSQP(algebra='builtin')
        osqp_solver.setup(
            P,
            q,
            A,
            l,
            u,
            eps_abs=eps_abs,
            eps_rel=eps_rel,
            max_iter=max_iter,
            verbose=False,
        )
        res = osqp_solver.solve()
        if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem not solved!')
        x1 = res.x
        y1 = res.y

        eps = grad_precision
        osqp_solver.setup(
            P,
            q + eps * dq,
            A,
            l,
            u,
            eps_abs=eps_abs,
            eps_rel=eps_rel,
            max_iter=max_iter,
            verbose=False,
        )
        res = osqp_solver.solve()
        if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem not solved!')
        x2 = res.x
        y2 = res.y

        dx_fd = (x2 - x1) / eps
        dy_fd = (y2 - y1) / eps
        dyl_fd = np.zeros(m)
        dyl_fd[y1 < 0] = -dy_fd[y1 < 0]
        dyu_fd = np.zeros(m)
        dyu_fd[y1 >= 0] = dy_fd[y1 >= 0]

        if verbose:
            print('dx_fd: ', np.round(dx_fd, decimals=4))
            print('dx: ', np.round(dx_computed, decimals=4))

        npt.assert_allclose(dx_fd, dx_computed, rtol=rel_tol, atol=abs_tol)
        npt.assert_allclose(dyl_fd, dyl_computed, rtol=rel_tol, atol=abs_tol)
        npt.assert_allclose(dyu_fd, dyu_computed, rtol=rel_tol, atol=abs_tol)

    @pytest.mark.skip(reason='forward derivatives not implemented yet')
    def test_eq_inf_forward(self, verbose=False):
        n, m = 10, 10

        prob = self.get_prob(n=n, m=m, P_scale=100.0, A_scale=100.0)
        P, q, A, l, u, true_x, true_y = prob
        l[:5] = u[:5]
        l[5:] = -osqp.constant('OSQP_INFTY', algebra='builtin')

        def grad(dq):
            [dx, dyl, dyu] = self.get_forward_grads(P, q, A, l, u, None, dq, None, None, None)
            return dx, dyl, dyu

        dq = np.random.normal(size=(n))
        dx_computed, dyl_computed, dyu_computed = grad(dq)
        osqp_solver = osqp.OSQP(algebra='builtin')
        osqp_solver.setup(
            P,
            q,
            A,
            l,
            u,
            eps_abs=eps_abs,
            eps_rel=eps_rel,
            max_iter=max_iter,
            verbose=False,
        )
        res = osqp_solver.solve()
        if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem not solved!')
        x1 = res.x
        y1 = res.y

        eps = grad_precision
        osqp_solver.setup(
            P,
            q + eps * dq,
            A,
            l,
            u,
            eps_abs=eps_abs,
            eps_rel=eps_rel,
            max_iter=max_iter,
            verbose=False,
        )
        res = osqp_solver.solve()
        if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem not solved!')
        x2 = res.x
        y2 = res.y

        dx_fd = (x2 - x1) / eps
        dy_fd = (y2 - y1) / eps
        dyl_fd = np.zeros(m)
        dyl_fd[y1 < 0] = -dy_fd[y1 < 0]
        dyu_fd = np.zeros(m)
        dyu_fd[y1 >= 0] = dy_fd[y1 >= 0]

        if verbose:
            print('dx_fd: ', np.round(dx_fd, decimals=4))
            print('dx: ', np.round(dx_computed, decimals=4))

        npt.assert_allclose(dx_fd, dx_computed, rtol=rel_tol, atol=abs_tol)
        npt.assert_allclose(dyl_fd, dyl_computed, rtol=rel_tol, atol=abs_tol)
        npt.assert_allclose(dyu_fd, dyu_computed, rtol=rel_tol, atol=abs_tol)

    @pytest.mark.skip(reason='forward derivatives not implemented yet')
    def test_multiple_forward_derivative(self, verbose=False):
        n, m = 5, 5

        prob = self.get_prob(n=n, m=m, P_scale=100.0, A_scale=100.0)
        P, q, A, l, u, true_x, true_y = prob

        def grad(dP, dq, dA, dl, du):
            [dx, dyl, dyu] = self.get_forward_grads(P, q, A, l, u, dP, dq, dA, dl, du)
            return dx, dyl, dyu

        dq = np.random.normal(size=(n))
        dA = sparse.csc_matrix(np.random.normal(size=(m, n)))
        dl = np.random.normal(size=(m))
        du = np.random.normal(size=(m))
        dL = np.random.normal(size=(n, n))
        dP = dL + dL.T
        dx_computed, dyl_computed, dyu_computed = grad(dP, dq, dA, dl, du)
        osqp_solver = osqp.OSQP(algebra='builtin')
        osqp_solver.setup(
            P,
            q,
            A,
            l,
            u,
            eps_abs=eps_abs,
            eps_rel=eps_rel,
            max_iter=max_iter,
            verbose=False,
        )
        res = osqp_solver.solve()
        if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem not solved!')
        x1 = res.x
        y1 = res.y

        eps = grad_precision
        osqp_solver.setup(
            P + eps * dP,
            q + eps * dq,
            A + eps * dA,
            l + eps * dl,
            u + eps * du,
            eps_abs=eps_abs,
            eps_rel=eps_rel,
            max_iter=max_iter,
            verbose=False,
        )
        res = osqp_solver.solve()
        if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem not solved!')
        x2 = res.x
        y2 = res.y

        dx_fd = (x2 - x1) / eps
        dy_fd = (y2 - y1) / eps
        dyl_fd = np.zeros(m)
        dyl_fd[y1 < 0] = -dy_fd[y1 < 0]
        dyu_fd = np.zeros(m)
        dyu_fd[y1 >= 0] = dy_fd[y1 >= 0]

        if verbose:
            print('dx_fd: ', np.round(dx_fd, decimals=4))
            print('dx: ', np.round(dx_computed, decimals=4))

        npt.assert_allclose(dx_fd, dx_computed, rtol=rel_tol, atol=abs_tol)
        npt.assert_allclose(dyl_fd, dyl_computed, rtol=rel_tol, atol=abs_tol)
        npt.assert_allclose(dyu_fd, dyu_computed, rtol=rel_tol, atol=abs_tol)

    def test_dl_dq(self, verbose=False):
        n, m = 5, 5

        prob = self.get_prob(n=n, m=m, P_scale=100.0, A_scale=100.0)
        P, q, A, l, u, true_x, true_y = prob

        def grad(q):
            dP, dq, _, _, _ = self.get_grads(P, q, A, l, u, true_x)
            return dq

        def f(q):
            m = osqp.OSQP(algebra='builtin')
            m.setup(
                P,
                q,
                A,
                l,
                u,
                eps_abs=eps_abs,
                eps_rel=eps_rel,
                max_iter=max_iter,
                verbose=False,
            )
            res = m.solve()
            if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
                raise ValueError('Problem not solved!')
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dq = grad(q)
        dq_fd = approx_fprime(q, f, grad_precision)

        if verbose:
            print('dq_fd: ', np.round(dq_fd, decimals=4))
            print('dq: ', np.round(dq, decimals=4))

        npt.assert_allclose(dq_fd, dq, rtol=rel_tol, atol=abs_tol)

    def test_dl_dP(self, verbose=False):
        n, m = 3, 3

        prob = self.get_prob(n=n, m=m, P_scale=100.0, A_scale=100.0)
        P, q, A, l, u, true_x, true_y = prob
        P_idx = P.nonzero()

        def grad(P_val):
            P_qp = sparse.csc_matrix((P_val, P_idx), shape=P.shape)
            dP, _, _, _, _ = self.get_grads(P_qp, q, A, l, u, true_x)
            return dP

        def f(P_val):
            P_qp = sparse.csc_matrix((P_val, P_idx), shape=P.shape)
            m = osqp.OSQP(algebra='builtin')
            m.setup(
                P_qp,
                q,
                A,
                l,
                u,
                eps_abs=eps_abs,
                eps_rel=eps_rel,
                max_iter=max_iter,
                verbose=False,
            )
            res = m.solve()
            if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
                raise ValueError('Problem not solved!')
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dP = grad(P.data)
        dP_fd_val = approx_fprime(P.data, f, grad_precision)
        dP_fd = sparse.csc_matrix((dP_fd_val, P_idx), shape=P.shape)
        dP_fd = (dP_fd + dP_fd.T) / 2

        if verbose:
            print('dP_fd: ', np.round(dP_fd.data, decimals=4))
            print('dA: ', np.round(dP.data, decimals=4))

        npt.assert_allclose(np.triu(dP), np.triu(dP_fd.todense()), rtol=rel_tol, atol=abs_tol)

    def test_dl_dA(self, verbose=False):
        n, m = 3, 3

        prob = self.get_prob(n=n, m=m, P_scale=100.0, A_scale=100.0)
        P, q, A, l, u, true_x, true_y = prob
        A_idx = A.nonzero()

        def grad(A_val):
            A_qp = sparse.csc_matrix((A_val, A_idx), shape=A.shape)
            _, _, dA, _, _ = self.get_grads(P, q, A_qp, l, u, true_x)
            return dA

        def f(A_val):
            A_qp = sparse.csc_matrix((A_val, A_idx), shape=A.shape)
            m = osqp.OSQP(algebra='builtin')
            m.setup(
                P,
                q,
                A_qp,
                l,
                u,
                eps_abs=eps_abs,
                eps_rel=eps_rel,
                max_iter=max_iter,
                verbose=False,
            )
            res = m.solve()
            if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
                raise ValueError('Problem not solved!')
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dA = grad(A.data)
        dA_fd_val = approx_fprime(A.data, f, grad_precision)
        dA_fd = sparse.csc_matrix((dA_fd_val, A_idx), shape=A.shape)

        if verbose:
            print('dA_fd: ', np.round(dA_fd.data, decimals=4))
            print('dA: ', np.round(dA.data, decimals=4))

        npt.assert_allclose(dA, dA_fd.todense(), rtol=rel_tol, atol=abs_tol)

    def test_dl_dl(self, verbose=False):
        n, m = 30, 30

        prob = self.get_prob(n=n, m=m, P_scale=100.0, A_scale=100.0)
        P, q, A, l, u, true_x, true_y = prob

        def grad(l):
            _, _, _, dl, _ = self.get_grads(P, q, A, l, u, true_x)
            return dl

        def f(l):
            m = osqp.OSQP(algebra='builtin')
            m.setup(
                P,
                q,
                A,
                l,
                u,
                eps_abs=eps_abs,
                eps_rel=eps_rel,
                max_iter=max_iter,
                verbose=False,
            )
            res = m.solve()
            if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
                raise ValueError('Problem not solved!')
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dl_computed = grad(l)
        dl_fd = approx_fprime(l, f, grad_precision)

        if verbose:
            print('dl_fd: ', np.round(dl_fd, decimals=4).tolist())
            print('dl_computed: ', np.round(dl_computed, decimals=4).tolist())

        npt.assert_allclose(dl_fd, dl_computed, rtol=rel_tol, atol=abs_tol)

    def test_dl_du(self, verbose=False):
        n, m = 10, 20

        prob = self.get_prob(n=n, m=m, P_scale=100.0, A_scale=100.0)
        P, q, A, l, u, true_x, true_y = prob

        def grad(u):
            _, _, _, _, du = self.get_grads(P, q, A, l, u, true_x)
            return du

        def f(u):
            m = osqp.OSQP(algebra='builtin')
            m.setup(
                P,
                q,
                A,
                l,
                u,
                eps_abs=eps_abs,
                eps_rel=eps_rel,
                max_iter=max_iter,
                verbose=False,
            )
            res = m.solve()
            if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
                raise ValueError('Problem not solved!')
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        du_computed = grad(u)
        du_fd = approx_fprime(u, f, grad_precision)

        if verbose:
            print('du_fd: ', np.round(du_fd, decimals=4))
            print('du: ', np.round(du_computed, decimals=4))

        npt.assert_allclose(du_fd, du_computed, rtol=rel_tol, atol=abs_tol)

    def test_dl_dA_eq(self, verbose=False):
        n, m = 30, 20

        prob = self.get_prob(n=n, m=m, P_scale=100.0, A_scale=100.0)
        P, q, A, l, u, true_x, true_y = prob
        # u = l
        # l[10:20] = -osqp.constant('OSQP_INFTY', algebra='builtin')
        u[:10] = l[:10]

        A_idx = A.nonzero()

        def grad(A_val):
            A_qp = sparse.csc_matrix((A_val, A_idx), shape=A.shape)
            _, _, dA, _, _ = self.get_grads(P, q, A_qp, l, u, true_x)
            return dA

        def f(A_val):
            A_qp = sparse.csc_matrix((A_val, A_idx), shape=A.shape)
            m = osqp.OSQP(algebra='builtin')
            m.setup(
                P,
                q,
                A_qp,
                l,
                u,
                eps_abs=eps_abs,
                eps_rel=eps_rel,
                max_iter=max_iter,
                verbose=False,
            )
            res = m.solve()
            if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
                raise ValueError('Problem not solved!')
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dA_computed = grad(A.data)
        dA_fd_val = approx_fprime(A.data, f, grad_precision)
        dA_fd = sparse.csc_matrix((dA_fd_val, A_idx), shape=A.shape)

        if verbose:
            print('dA_fd: ', np.round(dA_fd.data, decimals=6))
            print('dA_computed: ', np.round(dA_computed.data, decimals=6))

        npt.assert_allclose(dA_computed, dA_fd.todense(), rtol=rel_tol, atol=abs_tol)

    def test_dl_dq_eq(self, verbose=False):
        n, m = 20, 15

        prob = self.get_prob(n=n, m=m, P_scale=1.0, A_scale=1.0)
        P, q, A, l, u, true_x, true_y = prob
        # u = l
        # l[20:40] = -osqp.constant('OSQP_INFTY', algebra='builtin')
        u[:20] = l[:20]

        def grad(q):
            _, dq, _, _, _ = self.get_grads(P, q, A, l, u, true_x)
            return dq

        def f(q):
            m = osqp.OSQP(algebra='builtin')
            m.setup(
                P,
                q,
                A,
                l,
                u,
                eps_abs=eps_abs,
                eps_rel=eps_rel,
                max_iter=max_iter,
                verbose=False,
            )
            res = m.solve()
            if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
                raise ValueError('Problem not solved!')
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dq_computed = grad(q)
        dq_fd = approx_fprime(q, f, grad_precision)

        if verbose:
            print('dq_fd: ', np.round(dq_fd, decimals=4))
            print('dq_computed: ', np.round(dq_computed, decimals=4))

        npt.assert_allclose(dq_fd, dq_computed, rtol=rel_tol, atol=abs_tol)

    def test_dl_dq_eq_large(self, verbose=False):
        n, m = 100, 120

        prob = self.get_prob(n=n, m=m, P_scale=1.0, A_scale=1.0)
        P, q, A, l, u, true_x, true_y = prob

        l[20:40] = -osqp.constant('OSQP_INFTY', algebra='builtin')
        u[:20] = l[:20]

        def grad(q):
            _, dq, _, _, _ = self.get_grads(P, q, A, l, u, true_x)
            return dq

        def f(q):
            m = osqp.OSQP(algebra='builtin')
            m.setup(
                P,
                q,
                A,
                l,
                u,
                eps_abs=eps_abs,
                eps_rel=eps_rel,
                max_iter=max_iter,
                verbose=False,
            )
            res = m.solve()
            if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
                raise ValueError('Problem not solved!')
            x_hat = res.x

            return 0.5 * np.sum(np.square(x_hat - true_x))

        dq_computed = grad(q)
        dq_fd = approx_fprime(q, f, grad_precision)

        if verbose:
            print('dq_fd: ', np.round(dq_fd, decimals=4))
            print('dq_computed: ', np.round(dq_computed, decimals=4))

        npt.assert_allclose(dq_fd, dq_computed, rtol=rel_tol_relaxed, atol=abs_tol_relaxed)

    def _test_dl_dq_nonzero_dy(self, verbose=False):
        n, m = 6, 3

        prob = self.get_prob(n=n, m=m, P_scale=1.0, A_scale=1.0)
        P, q, A, l, u, true_x, true_y = prob
        # u = l
        # l[20:40] = -osqp.constant('OSQP_INFTY', algebra='builtin')
        num_eq = 2
        u[:num_eq] = l[:num_eq]

        def grad(q):
            _, dq, _, _, _ = self.get_grads(P, q, A, l, u, true_x, true_y)
            return dq

        def f(q):
            m = osqp.OSQP(algebra='builtin')
            m.setup(
                P,
                q,
                A,
                l,
                u,
                eps_abs=eps_abs,
                eps_rel=eps_rel,
                max_iter=max_iter,
                verbose=False,
            )
            res = m.solve()
            if res.info.status_val != osqp.SolverStatus.OSQP_SOLVED:
                raise ValueError('Problem not solved!')
            x_hat = res.x
            y_hat = res.y
            yu_hat = np.maximum(y_hat, 0)
            yl_hat = -np.minimum(y_hat, 0)

            true_yu = np.maximum(true_y, 0)
            true_yl = -np.minimum(true_y, 0)
            # return 0.5 * np.sum(np.square(x_hat - true_x)) + np.sum(yl_hat) + np.sum(yu_hat)
            return 0.5 * (
                np.sum(np.square(x_hat - true_x))
                + np.sum(np.square(yl_hat - true_yl))
                + np.sum(np.square(yu_hat - true_yu))
            )

        dq_computed = grad(q)
        dq_fd = approx_fprime(q, f, grad_precision)

        if verbose:
            print('dq_fd: ', np.round(dq_fd, decimals=4))
            print('dq_computed: ', np.round(dq_computed, decimals=4))

        npt.assert_allclose(dq_fd, dq_computed, rtol=rel_tol, atol=abs_tol)



from types import SimpleNamespace
import osqp
import numpy as np
from scipy import sparse
import pytest
import numpy.testing as nptest
from osqp.tests.utils import load_high_accuracy


@pytest.fixture
def self(algebra, solver_type, atol, rtol, decimal_tol):
    self = SimpleNamespace()

    np.random.seed(4)

    self.n = 30
    self.m = 30
    self.P = sparse.csc_matrix((self.n, self.n))
    self.q = np.zeros(self.n)
    self.A = sparse.random(self.m, self.n, density=1.0, format='csc')
    self.u = np.random.rand(self.m)
    self.l = self.u
    self.opts = {
        'verbose': False,
        'eps_abs': 1e-06,
        'eps_rel': 1e-06,
        'scaling': True,
        'alpha': 1.6,
        'max_iter': 5000,
        'polishing': False,
        'warm_starting': True,
        'polish_refine_iter': 4,
        'solver_type': solver_type,
    }

    self.model = osqp.OSQP(algebra=algebra)
    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)

    self.rtol = rtol
    self.atol = atol
    self.decimal_tol = decimal_tol

    return self


def test_feasibility_problem(self):
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_feasibility_problem')

    if self.model.solver_type == 'direct':  # pytest-todo
        nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)
    else:
        assert res.info.status_val == self.model.constant('OSQP_MAX_ITER_REACHED')



from types import SimpleNamespace
import numpy as np
from scipy import sparse
import pytest
import numpy.testing as nptest
import osqp
from osqp.tests.utils import load_high_accuracy


@pytest.fixture
def self(algebra, solver_type, atol, rtol, decimal_tol):
    ns = SimpleNamespace()
    ns.opts = {
        'verbose': False,
        'eps_abs': 1e-03,
        'eps_rel': 1e-03,
        'scaling': True,
        'rho': 0.1,
        'alpha': 1.6,
        'max_iter': 2500,
        'polishing': True,
        'polish_refine_iter': 4,
        'solver_type': solver_type,
    }
    ns.model = osqp.OSQP(algebra=algebra)
    ns.atol = atol
    ns.rtol = rtol
    ns.decimal_tol = decimal_tol
    return ns


def test_polish_simple(self):
    # Simple QP problem
    self.P = sparse.diags([11.0, 0.0], format='csc')
    self.q = np.array([3, 4])
    self.A = sparse.csc_matrix([[-1, 0], [0, -1], [-1, -3], [2, 5], [3, 4]])
    self.u = np.array([0, 0, -15, 100, 80])
    self.l = -1e05 * np.ones(len(self.u))
    self.n = self.P.shape[0]
    self.m = self.A.shape[0]
    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)

    # Solve problem
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_polish_simple')
    # Assert close
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_polish_unconstrained(self):
    # Unconstrained QP problem
    np.random.seed(4)

    self.n = 30
    self.m = 0
    P = sparse.diags(np.random.rand(self.n)) + 0.2 * sparse.eye(self.n)
    self.P = P.tocsc()
    self.q = np.random.randn(self.n)
    self.A = sparse.csc_matrix((self.m, self.n))
    self.l = np.array([])
    self.u = np.array([])
    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)

    # Solve problem
    res = self.model.solve()

    x_sol, _, obj_sol = load_high_accuracy('test_polish_unconstrained')
    # Assert close
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_polish_random(self):
    # Random QP problem
    np.random.seed(6)

    self.n = 30
    self.m = 50
    Pt = sparse.random(self.n, self.n)
    self.P = Pt.T @ Pt
    self.q = np.random.randn(self.n)
    self.A = sparse.csc_matrix(np.random.randn(self.m, self.n))
    self.l = -3 + np.random.randn(self.m)
    self.u = 3 + np.random.randn(self.m)
    model = osqp.OSQP(algebra=self.model.algebra)
    model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)
    assert model.solver_type == self.opts['solver_type']

    # Solve problem
    res = model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_polish_random')
    # Assert close
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)



import sys
import os
from types import SimpleNamespace
from enum import IntEnum
import shutil
import subprocess
import warnings
import importlib
import importlib.resources
import numpy as np
import scipy.sparse as spa
from jinja2 import Environment, PackageLoader, select_autoescape

_ALGEBRAS = (
    'cuda',
    'mkl',
    'builtin',
)  # Highest->Lowest priority of algebras that are tried in turn
# Mapping from algebra to loadable module
_ALGEBRA_MODULES = {
    'cuda': 'osqp_cuda',
    'mkl': 'osqp_mkl',
    'builtin': 'osqp.ext_builtin',
}
OSQP_ALGEBRA_BACKEND = os.environ.get('OSQP_ALGEBRA_BACKEND')  # If envvar is set, that algebra is used by default


def algebra_available(algebra):
    assert algebra in _ALGEBRAS, f'Unknown algebra {algebra}'
    module = _ALGEBRA_MODULES[algebra]

    try:
        importlib.import_module(module)
    except ImportError:
        return False
    else:
        return True


def algebras_available():
    return [algebra for algebra in _ALGEBRAS if algebra_available(algebra)]


def default_algebra():
    if OSQP_ALGEBRA_BACKEND is not None:
        return OSQP_ALGEBRA_BACKEND
    for algebra in _ALGEBRAS:
        if algebra_available(algebra):
            return algebra
    raise RuntimeError('No algebra backend available!')


def default_algebra_module():
    """
    Get the default algebra module.
    Note: importlib.import_module is cached so we pay almost no penalty
      for repeated calls to this function.
    """
    return importlib.import_module(_ALGEBRA_MODULES[default_algebra()])


def constant(which, algebra='builtin'):
    """
    Get a named constant from the extension module.
    Since constants are typically consistent across osqp algebras,
    we use the `builtin` algebra (always guaranteed to be available)
    by default.
    """
    m = importlib.import_module(_ALGEBRA_MODULES[algebra])
    _constant = getattr(m, which, None)

    if which in m.osqp_status_type.__members__:
        warnings.warn(
            'Direct access to osqp status values will be deprecated. Please use the SolverStatus enum instead.',
            PendingDeprecationWarning,
        )

    # If the constant was exported directly as an atomic type in the extension, use it;
    # Otherwise it's an enum out of which we can obtain the raw value
    if isinstance(_constant, (int, float, str)):
        return _constant
    elif _constant is not None:
        return _constant.value
    else:
        # Handle special cases
        if which == 'OSQP_NAN':
            return np.nan

        raise RuntimeError(f'Unknown constant {which}')


def construct_enum(name, binding_enum_name):
    """
    Dynamically construct an IntEnum from available enum members.
    For all values, see https://osqp.org/docs/interfaces/status_values.html
    """
    m = default_algebra_module()
    binding_enum = getattr(m, binding_enum_name)
    return IntEnum(name, [(v.name, v.value) for v in binding_enum.__members__.values()])


SolverStatus = construct_enum('SolverStatus', 'osqp_status_type')
SolverError = construct_enum('SolverError', 'osqp_error_type')


class OSQPException(Exception):
    """
    OSQPException is raised by the wrapper interface when it encounters an
    exception by the underlying OSQP solver.
    """

    def __init__(self, error_code=None):
        if error_code:
            self.args = (error_code,)

    def __eq__(self, error_code):
        return len(self.args) > 0 and self.args[0] == error_code


class OSQP:

    """
    For OSQP bindings (see bindings.cpp.in) that throw `ValueError`s
    (through `throw py::value_error(...)`), we catch and re-raise them
    as `OSQPException`s, with the correct int value as args[0].
    """

    @classmethod
    def raises_error(cls, fn, *args, **kwargs):
        try:
            return_value = fn(*args, **kwargs)
        except ValueError as e:
            if e.args:
                error_code = None
                try:
                    error_code = int(e.args[0])
                except ValueError:
                    pass
            raise OSQPException(error_code)
        else:
            return return_value

    def __init__(self, *args, **kwargs):
        self.m = None
        self.n = None

        self.algebra = kwargs.pop('algebra') if 'algebra' in kwargs else default_algebra()
        if not algebra_available(self.algebra):
            raise RuntimeError(f'Algebra {self.algebra} not available')
        self.ext = importlib.import_module(_ALGEBRA_MODULES[self.algebra])

        self._dtype = np.float32 if self.ext.OSQP_USE_FLOAT == 1 else np.float64
        self._itype = np.int64 if self.ext.OSQP_USE_LONG == 1 else np.int32

        # The following attributes are populated on setup()
        self._solver = None
        self._derivative_cache = {}

    def __str__(self):
        if self._solver is None:
            return f'Uninitialized OSQP with algebra={self.algebra}'
        else:
            return f'OSQP with algebra={self.algebra} ({self.solver_type})'

    def _infer_mnpqalu(self, P=None, q=None, A=None, l=None, u=None):
        # infer as many parameters of the problems as we can, and return them as a tuple
        if P is None:
            if q is not None:
                n = len(q)
            elif A is not None:
                n = A.shape[1]
            else:
                raise ValueError('The problem does not have any variables')
        else:
            n = P.shape[0]

        m = 0 if A is None else A.shape[0]

        if A is None:
            assert (l is None) and (u is None), 'If A is unspecified, leave l/u unspecified too.'
        else:
            assert (l is not None) or (u is not None), 'If A is specified, specify at least one of l/u.'
            if l is None:
                l = -np.inf * np.ones(A.shape[0])
            if u is None:
                u = np.inf * np.ones(A.shape[0])

        if P is None:
            P = spa.csc_matrix(
                (
                    np.zeros((0,), dtype=self._dtype),  # data
                    np.zeros((0,), dtype=self._itype),  # indices
                    np.zeros((n + 1,), dtype=self._itype),
                ),  # indptr
                shape=(n, n),
            )
        if q is None:
            q = np.zeros(n)

        if A is None:
            A = spa.csc_matrix(
                (
                    np.zeros((0,), dtype=self._dtype),  # data
                    np.zeros((0,), dtype=self._itype),  # indices
                    np.zeros((n + 1,), dtype=self._itype),
                ),  # indptr
                shape=(m, n),
            )
            l = np.zeros(A.shape[0])
            u = np.zeros(A.shape[0])

        assert len(q) == n, 'Incorrect dimension of q'
        assert len(l) == m, 'Incorrect dimension of l'
        assert len(u) == m, 'Incorrect dimension of u'

        if not spa.issparse(P) and isinstance(P, np.ndarray) and P.ndim == 2:
            raise TypeError('P is required to be a sparse matrix')
        if not spa.issparse(A) and isinstance(A, np.ndarray) and A.ndim == 2:
            raise TypeError('A is required to be a sparse matrix')

        if spa.tril(P, -1).data.size > 0:
            P = spa.triu(P, format='csc')

        # Convert matrices in CSC form to individual pointers
        if not spa.isspmatrix_csc(P):
            warnings.warn('Converting sparse P to a CSC matrix. This may take a while...')
            P = P.tocsc()
        if not spa.isspmatrix_csc(A):
            warnings.warn('Converting sparse A to a CSC matrix. This may take a while...')
            A = A.tocsc()

        if not P.has_sorted_indices:
            P.sort_indices()
        if not A.has_sorted_indices:
            A.sort_indices()

        u = np.minimum(u, self.constant('OSQP_INFTY'))
        l = np.maximum(l, -self.constant('OSQP_INFTY'))

        return m, n, P, q, A, l, u

    @property
    def capabilities(self):
        return int(self.ext.osqp_capabilities())

    def has_capability(self, capability: str):
        try:
            cap = int(self.ext.osqp_capabilities_type.__members__[capability])
        except KeyError:
            raise RuntimeError(f'Unrecognized capability {capability}')

        return (self.capabilities & cap) != 0

    @property
    def solver_type(self):
        return (
            'direct'
            if self.settings.linsys_solver == self.ext.osqp_linsys_solver_type.OSQP_DIRECT_SOLVER
            else 'indirect'
        )

    @property
    def cg_preconditioner(self):
        return 'diagonal' if self.settings.cg_precond == self.ext.OSQP_DIAGONAL_PRECONDITIONER else None

    def _as_dense(self, m):
        assert isinstance(m, self.ext.CSC)
        _m_csc = spa.csc_matrix((m.x, m.i, m.p))
        return np.array(_m_csc.todense())

    def _csc_triu_as_csc_full(self, m):
        _m_triu_dense = self._as_dense(m)
        _m_full_dense = np.tril(_m_triu_dense.T, -1) + _m_triu_dense
        _m_full_csc = spa.csc_matrix(_m_full_dense)
        return self.ext.CSC(_m_full_csc)

    def constant(self, which):
        return constant(which, algebra=self.algebra)

    def update_settings(self, **kwargs):
        assert self.settings is not None

        # Some setting names have changed. Support the old names for now, but warn the caller.
        renamed_settings = {
            'polish': 'polishing',
            'warm_start': 'warm_starting',
        }
        for k, v in renamed_settings.items():
            if k in kwargs:
                warnings.warn(
                    f'"{k}" is deprecated. Please use "{v}" instead.',
                    DeprecationWarning,
                )
                kwargs[v] = kwargs[k]
                del kwargs[k]

        settings_changed = False

        if 'rho' in kwargs and self._solver is not None:
            self._solver.update_rho(kwargs.pop('rho'))
        if 'solver_type' in kwargs:
            value = kwargs.pop('solver_type')
            assert value in ('direct', 'indirect')
            self.settings.linsys_solver = (
                self.ext.osqp_linsys_solver_type.OSQP_DIRECT_SOLVER
                if value == 'direct'
                else self.ext.osqp_linsys_solver_type.OSQP_INDIRECT_SOLVER
            )
            settings_changed = True
        if 'cg_preconditioner' in kwargs:
            value = kwargs.pop('cg_preconditioner')
            assert value in (None, 'diagonal')
            self.settings.cg_precond = (
                self.ext.OSQP_DIAGONAL_PRECONDITIONER if value == 'diagonal' else self.ext.OSQP_NO_PRECONDITIONER
            )
            settings_changed = True

        for k in self.ext.OSQPSettings.__dict__:
            if not k.startswith('__'):
                if k in kwargs:
                    setattr(self.settings, k, kwargs.pop(k))
                    settings_changed = True

        if kwargs:
            raise ValueError(f'Unrecognized settings {list(kwargs.keys())}')

        if settings_changed and self._solver is not None:
            self.raises_error(self._solver.update_settings, self.settings)

    def update(self, **kwargs):
        # TODO: sanity-check on types/dimensions

        q, l, u = kwargs.get('q'), kwargs.get('l'), kwargs.get('u')
        if l is not None:
            l = np.maximum(l, -self.constant('OSQP_INFTY'))
        if u is not None:
            u = np.minimum(u, self.constant('OSQP_INFTY'))

        if q is not None or l is not None or u is not None:
            self._solver.update_data_vec(q=q, l=l, u=u)
        if 'Px' in kwargs or 'Px_idx' in kwargs or 'Ax' in kwargs or 'Ax_idx' in kwargs:
            self._solver.update_data_mat(
                P_x=kwargs.get('Px'),
                P_i=kwargs.get('Px_idx'),
                A_x=kwargs.get('Ax'),
                A_i=kwargs.get('Ax_idx'),
            )

        if q is not None:
            self._derivative_cache['q'] = q
        if l is not None:
            self._derivative_cache['l'] = l
        if u is not None:
            self._derivative_cache['u'] = u

        for _var in ('P', 'A'):
            _varx = f'{_var}x'
            if kwargs.get(_varx) is not None:
                if kwargs.get(f'{_varx}_idx') is None:
                    self._derivative_cache[_var].data = kwargs[_varx]
                else:
                    self._derivative_cache[_var].data[kwargs[f'{_varx}_idx']] = kwargs[_varx]

        # delete results from self._derivative_cache to prohibit
        # taking the derivative of unsolved problems
        self._derivative_cache.pop('results', None)
        self._derivative_cache.pop('solver', None)
        self._derivative_cache.pop('M', None)

    def setup(self, P, q, A, l, u, **settings):
        m, n, P, q, A, l, u = self._infer_mnpqalu(P=P, q=q, A=A, l=l, u=u)
        self._derivative_cache.update({'P': P, 'q': q, 'A': A, 'l': l, 'u': u})
        self.m = m
        self.n = n
        P = self.ext.CSC(P.astype(self._dtype))
        q = q.astype(self._dtype)
        A = self.ext.CSC(A.astype(self._dtype))
        l = l.astype(self._dtype)
        u = u.astype(self._dtype)

        self.settings = self.ext.OSQPSettings()
        self.ext.osqp_set_default_settings(self.settings)
        self.update_settings(**settings)

        self._solver = self.raises_error(
            self.ext.OSQPSolver,
            P,
            q,
            A,
            l,
            u,
            self.m,
            self.n,
            self.settings,
        )
        if 'rho' in settings:
            self._solver.update_rho(settings['rho'])

    def warm_start(self, x=None, y=None):
        # TODO: sanity checks on types/dimensions
        return self._solver.warm_start(x, y)

    def solve(self, raise_error=None):
        if raise_error is None:
            warnings.warn(
                'The default value of raise_error will change to True in the future.',
                PendingDeprecationWarning,
            )
            raise_error = False

        self._solver.solve()

        info = self._solver.info
        if info.status_val == SolverStatus.OSQP_NON_CVX:
            info.obj_val = np.nan

        if info.status_val != SolverStatus.OSQP_SOLVED and raise_error:
            raise OSQPException(info.status_val)

        # Create a Namespace of OSQPInfo keys and associated values
        _info = SimpleNamespace(**{k: getattr(info, k) for k in info.__class__.__dict__ if not k.startswith('__')})

        # TODO: The following structure is only to maintain backward compatibility, where x/y are attributes
        # directly inside the returned object on solve(). This should be simplified!
        results = SimpleNamespace(
            x=self._solver.solution.x,
            y=self._solver.solution.y,
            prim_inf_cert=self._solver.solution.prim_inf_cert,
            dual_inf_cert=self._solver.solution.dual_inf_cert,
            info=_info,
        )

        self._derivative_cache['results'] = results
        return results

    def _render_pywrapper_files(self, output_folder, **kwargs):
        env = Environment(
            loader=PackageLoader('osqp.codegen.pywrapper', package_path=''),
            autoescape=select_autoescape(),
        )

        for template_name in env.list_templates(extensions='.jinja'):
            template = env.get_template(template_name)
            template_base_name = os.path.splitext(template_name)[0]

            with open(os.path.join(output_folder, template_base_name), 'w') as f:
                f.write(template.render(**kwargs))

    def codegen(
        self,
        folder,
        parameters='vectors',
        extension_name='emosqp',
        force_rewrite=False,
        use_float=False,
        printing_enable=False,
        profiling_enable=False,
        interrupt_enable=False,
        include_codegen_src=True,
        prefix='',
        compile=False,
    ):
        assert self.has_capability('OSQP_CAPABILITY_CODEGEN'), 'This OSQP object does not support codegen'
        assert parameters in (
            'vectors',
            'matrices',
        ), 'Unknown parameters specification'

        defines = self.ext.OSQPCodegenDefines()
        self.ext.osqp_set_default_codegen_defines(defines)

        defines.embedded_mode = 1 if parameters == 'vectors' else 2
        defines.float_type = 1 if use_float else 0
        defines.printing_enable = 1 if printing_enable else 0
        defines.profiling_enable = 1 if profiling_enable else 0
        defines.interrupt_enable = 1 if interrupt_enable else 0
        defines.derivatives_enable = 0

        folder = os.path.abspath(folder)
        if include_codegen_src:
            # https://github.com/python/importlib_resources/issues/85
            try:
                codegen_src_path = importlib.resources.files('osqp.codegen').joinpath('codegen_src')
                shutil.copytree(codegen_src_path, folder, dirs_exist_ok=force_rewrite)
            except AttributeError:
                handle = importlib.resources.path('osqp.codegen', 'codegen_src')
                with handle as codegen_src_path:
                    shutil.copytree(codegen_src_path, folder, dirs_exist_ok=force_rewrite)

        # The C codegen call expects the folder to exist and have a trailing slash
        os.makedirs(folder, exist_ok=True)
        if not folder.endswith(os.path.sep):
            folder += os.path.sep

        status = self._solver.codegen(folder, prefix, defines)
        assert status == 0, f'Codegen failed with error code {status}'

        if extension_name is not None:
            assert include_codegen_src, 'If generating python wrappers, include_codegen_src must be True'
            template_vars = dict(
                prefix=prefix,
                extension_name=extension_name,
                embedded_mode=defines.embedded_mode,
            )
            self._render_pywrapper_files(folder, **template_vars)
            if compile:
                subprocess.check_call(
                    [
                        sys.executable,
                        'setup.py',
                        'build_ext',
                        '--inplace',
                    ],
                    cwd=folder,
                )

        return folder

    def adjoint_derivative_compute(self, dx=None, dy=None):
        """
        Compute adjoint derivative after solve.
        """

        assert self.has_capability('OSQP_CAPABILITY_DERIVATIVES'), 'This OSQP object does not support derivatives'

        try:
            results = self._derivative_cache['results']
        except KeyError:
            raise ValueError(
                'Problem has not been solved. ' 'You cannot take derivatives. ' 'Please call the solve function.'
            )

        if results.info.status_val != SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem has not been solved to optimality. ' 'You cannot take derivatives')

        if dy is None:
            dy = np.zeros(self.m)

        self._solver.adjoint_derivative_compute(dx, dy)

    def adjoint_derivative_get_mat(self, as_dense=True, dP_as_triu=True):
        """
        Get dP/dA matrices after an invocation of adjoint_derivative_compute
        """

        assert self.has_capability('OSQP_CAPABILITY_DERIVATIVES'), 'This OSQP object does not support derivatives'

        try:
            results = self._derivative_cache['results']
        except KeyError:
            raise ValueError(
                'Problem has not been solved. ' 'You cannot take derivatives. ' 'Please call the solve function.'
            )

        if results.info.status_val != SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem has not been solved to optimality. ' 'You cannot take derivatives')

        P, _ = self._derivative_cache['P'], self._derivative_cache['q']
        A = self._derivative_cache['A']

        dP = self.ext.CSC(P.copy())
        dA = self.ext.CSC(A.copy())

        self._solver.adjoint_derivative_get_mat(dP, dA)

        if not dP_as_triu:
            dP = self._csc_triu_as_csc_full(dP)

        if as_dense:
            dP = self._as_dense(dP)
            dA = self._as_dense(dA)

        return dP, dA

    def adjoint_derivative_get_vec(self):
        """
        Get dq/dl/du vectors after an invocation of adjoint_derivative_compute
        """

        assert self.has_capability('OSQP_CAPABILITY_DERIVATIVES'), 'This OSQP object does not support derivatives'

        try:
            results = self._derivative_cache['results']
        except KeyError:
            raise ValueError(
                'Problem has not been solved. ' 'You cannot take derivatives. ' 'Please call the solve function.'
            )

        if results.info.status_val != SolverStatus.OSQP_SOLVED:
            raise ValueError('Problem has not been solved to optimality. ' 'You cannot take derivatives')

        dq = np.empty(self.n).astype(self._dtype)
        dl = np.zeros(self.m).astype(self._dtype)
        du = np.zeros(self.m).astype(self._dtype)

        self._solver.adjoint_derivative_get_vec(dq, dl, du)

        return dq, dl, du



import os.path
import numpy as np


def load_high_accuracy(test_name):
    npz = os.path.join(os.path.dirname(__file__), 'solutions', f'{test_name}.npz')
    npzfile = np.load(npz)
    return npzfile['x_val'], npzfile['y_val'], npzfile['obj']



from types import SimpleNamespace
import numpy as np
from scipy import sparse
import pytest
import osqp


@pytest.fixture
def self(algebra, solver_type, atol, rtol, decimal_tol):
    ns = SimpleNamespace()
    ns.opts = {
        'verbose': True,
        'adaptive_rho': False,
        'eps_abs': 1e-08 if solver_type == 'direct' else 1e-2,
        'eps_rel': 1e-08 if solver_type == 'direct' else 1e-2,
        'polishing': False,
        'check_termination': 1,
        'solver_type': solver_type,
    }

    ns.model = osqp.OSQP(algebra=algebra)
    return ns


def test_warm_start(self):
    # Big problem
    np.random.seed(2)
    self.n = 100
    self.m = 200
    self.A = sparse.random(self.m, self.n, density=0.9, format='csc')
    self.l = -np.random.rand(self.m) * 2.0
    self.u = np.random.rand(self.m) * 2.0

    P = sparse.random(self.n, self.n, density=0.9)
    self.P = sparse.triu(P.dot(P.T), format='csc')
    self.q = np.random.randn(self.n)

    # Setup solver
    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, **self.opts)

    # Solve problem with OSQP
    res = self.model.solve()

    # Store optimal values
    x_opt = res.x
    y_opt = res.y
    tot_iter = res.info.iter

    # Warm start with zeros and check if number of iterations is the same
    self.model.warm_start(x=np.zeros(self.n), y=np.zeros(self.m))
    res = self.model.solve()
    assert res.info.iter == tot_iter

    # Warm start with optimal values and check that number of iter < 10
    self.model.warm_start(x=x_opt, y=y_opt)
    res = self.model.solve()
    assert res.info.iter < 10
